
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <ctype.h>
#include <errno.h>

#include "errors.h"

typedef struct
{
  int  id;
  char text[128];
} ErrorText;

static ErrorText errorTexts[63];
static int       errorTextCount = 0;
static int       errorTextId    = 0;

int _Errors_textToIndex(const char *text)
{
  int index;
  int minId;
  int z,i;

  errorTextId++;
  if (errorTextCount < 63)
  {
    index = errorTextCount;
    errorTextCount++;
  }
  else
  {
    index = 0;
    minId = INT_MAX;
    for (z = 0; z < 63; z++)
    {
      if (errorTexts[z].id < minId)
      {
        index = z;
        minId = errorTexts[z].id;
      }
    }
  }
  z = 0;
  i = 0;
  while ((z < strlen(text)) && (i < 128-1-1))
  {
    if (!iscntrl(text[z])) { errorTexts[index].text[i] = text[z]; i++; }
    z++;
  }
  errorTexts[index].text[i] = '\0';
  errorTexts[errorTextCount].id = errorTextId;
  return index+1;
}

#include "config.h"
#ifdef HAVE_LIBSSH2
#include "libssh2.h"
#endif

#define GET_ERROR_CODE(error)      (((error) & 0x000003FF) >> 0)
#define GET_ERROR_TEXTINDEX(error) (((error) & 0x0000FC00) >> 10)
#define GET_ERROR_TEXT(error)      ((GET_ERROR_TEXTINDEX(error)>0)?errorTexts[GET_ERROR_TEXTINDEX(error)-1].text:"unknown")
#define GET_ERRNO(error)           ((long)((error) & 0xFFFF0000) >> 16)

#define ERROR_CPDE GET_ERROR_CODE(error)
#define ERROR_TEXT GET_ERROR_TEXT(error)
#define ERRNO      GET_ERRNO(error)

unsigned int Errors_getCode(Errors error)
{
  return GET_ERROR_CODE(error);
}

const char *Errors_getText(Errors error)
{
  static char errorText[128];

  strcpy(errorText,"unknown");
  switch (GET_ERROR_CODE(error))
  {
    case ERROR_NONE: return "none";
    case ERROR_INSUFFICIENT_MEMORY: return "insufficient memory";
    case ERROR_INIT: return "init fail";
    case ERROR_INVALID_ARGUMENT: return "invalid argument";
    case ERROR_CONFIG: return "config error";
    case ERROR_ABORTED: return "aborted";
    case ERROR_FUNCTION_NOT_SUPPORTED: return "function not supported";
    case ERROR_STILL_NOT_IMPLEMENTED: return "function still not implemented";
    case ERROR_DAEMON_FAIL: return "run as daemon fail";
    case ERROR_INVALID_PATTERN: return "init pattern matching";
    case ERROR_INIT_TLS: return "init TLS (SSL) fail";
    case ERROR_NO_TLS_CA: return "no TLS (SSL) certificate authority file 'bar-ca.pem'";
    case ERROR_NO_TLS_CERTIFICATE: return "no TLS (SSL) certificate file 'bar-server-cert.pem'";
    case ERROR_NO_TLS_KEY: return "no or unreadable TLS (SSL) key file 'bar-server-key.pem'";
    case ERROR_INVALID_TLS_CA: return "invalid TLS (SSL) certificate authority";
    case ERROR_INVALID_TLS_CERTIFICATE: return "invalid TLS (SSL) certificate";
    case ERROR_TLS_CERTIFICATE_EXPIRED: return "TLS (SSL) certificate expired";
    case ERROR_TLS_CERTIFICATE_NOT_ACTIVE: return "TLS (SSL) certificate is still not active";
    case ERROR_TLS_HANDSHAKE: return "TLS (SSL) handshake failure";
    case ERROR_INVALID_SSH_SPECIFIER: return "invalid SSH specifier";
    case ERROR_NO_SSH_PASSWORD: return "no SSH password given";
    case ERROR_INVALID_SSH_PASSWORD: return "invalid SSH password";
    case ERROR_SSH_SESSION_FAIL: return "initialize SSH session fail";
    case ERROR_SSH_AUTHENTIFICATION:
      {
        strncpy(errorText,"SSH authentification fail",sizeof(errorText));
        if (ERROR_TEXT != NULL)
        {
          strncat(errorText,": ",sizeof(errorText)-1-strlen(errorText));
          strncat(errorText,ERROR_TEXT,sizeof(errorText)-1-strlen(errorText));
        }
      }
      break;
    case ERROR_SSH:
      {
        strncpy(errorText,"ssh protocol error",sizeof(errorText));
        if (ERROR_TEXT != NULL)
        {
          strncat(errorText,": ",sizeof(errorText)-1-strlen(errorText));
          strncat(errorText,ERROR_TEXT,sizeof(errorText)-1-strlen(errorText));
        }
      }
      break;
    case ERROR_INVALID_FTP_SPECIFIER: return "invalid FTP specifier";
    case ERROR_FTP_SESSION_FAIL: return "initialize FTP session fail";
    case ERROR_NO_FTP_PASSWORD: return "no FTP password given";
    case ERROR_INVALID_FTP_PASSWORD: return "invalid FTP password";
    case ERROR_FTP_AUTHENTIFICATION: return "invalid FTP user/password";
    case ERROR_INIT_COMPRESS: return "init compress fail";
    case ERROR_COMPRESS_ERROR: return "compress fail";
    case ERROR_DEFLATE_ERROR: return "deflate fail";
    case ERROR_INFLATE_ERROR: return "inflate fail";
    case ERROR_COMPRESS_EOF: return "end of compressed file";
    case ERROR_UNSUPPORTED_BLOCK_SIZE: return "unsupported block size";
    case ERROR_INIT_CRYPT: return "init crypt fail";
    case ERROR_NO_CRYPT_PASSWORD: return "no password given for cipher";
    case ERROR_INVALID_PASSWORD: return "invalid password";
    case ERROR_PASSWORDS_MISMATCH: return "passwords mismatch";
    case ERROR_INIT_CIPHER: return "init cipher fail";
    case ERROR_ENCRYPT_FAIL: return "encrypt fail";
    case ERROR_DECRYPT_FAIL: return "decrypt fail";
    case ERROR_CREATE_KEY_FAIL: return ERROR_TEXT;
    case ERROR_INVALID_KEY: return "invalid key";
    case ERROR_INVALID_KEY_LENGTH: return "invalid key length";
    case ERROR_INVALID_BLOCK_LENGTH: return "invalid block length";
    case ERROR_NO_PUBLIC_KEY: return "no public key";
    case ERROR_NO_PRIVATE_KEY: return "no private key";
    case ERROR_NOT_A_PUBLIC_KEY: return "key is not a public key";
    case ERROR_NOT_A_PRIVATE_KEY: return "key is not a private key";
    case ERROR_KEY_ENCRYPT_FAIL: return "key encryption fail";
    case ERROR_KEY_DECRYPT_FAIL: return "key decryption fail";
    case ERROR_WRONG_PRIVATE_KEY: return "wrong private key";
    case ERROR_CREATE_FILE:
    case ERROR_OPEN_FILE:
    case ERROR_OPEN_DIRECTORY:
      {
        strncpy(errorText,strerror(ERRNO),sizeof(errorText)-1); errorText[sizeof(errorText)-1] = '\0';
        if (ERROR_TEXT != NULL)
        {
          strncat(errorText,": ",sizeof(errorText)-1-strlen(errorText));
          strncat(errorText,ERROR_TEXT,sizeof(errorText)-1-strlen(errorText));
        }
      }
      break;
    case ERROR_IO_ERROR:
      {
        strncpy(errorText,strerror(ERRNO),sizeof(errorText)-1); errorText[sizeof(errorText)-1] = '\0';
        if (ERROR_TEXT != NULL)
        {
          strncat(errorText,": ",sizeof(errorText)-1-strlen(errorText));
          strncat(errorText,ERROR_TEXT,sizeof(errorText)-1-strlen(errorText));
        }
      }
      break;
    case ERROR_READ_FILE: return "read file fail";
    case ERROR_WRITE_FILE: return "write file fail";
    case ERROR_DELETE_FILE: return "delete file fail";
    case ERROR_FILE_EXITS: return "file already exists";
    case ERROR_FILE_NOT_FOUND: return "file not found";
    case ERROR_NOT_A_DIRECTORY:
      {
        strncpy(errorText,"not a directory",sizeof(errorText)-1); errorText[sizeof(errorText)-1] = '\0';
        if (ERROR_TEXT != NULL)
        {
          strncat(errorText,": ",sizeof(errorText)-1-strlen(errorText));
          strncat(errorText,ERROR_TEXT,sizeof(errorText)-1-strlen(errorText));
        }
      }
      break;
    case ERROR_OPEN_DEVICE:
      {
        strncpy(errorText,strerror(ERRNO),sizeof(errorText)-1); errorText[sizeof(errorText)-1] = '\0';
        if (ERROR_TEXT != NULL)
        {
          strncat(errorText,": ",sizeof(errorText)-1-strlen(errorText));
          strncat(errorText,ERROR_TEXT,sizeof(errorText)-1-strlen(errorText));
        }
      }
      break;
    case ERROR_INVALID_DEVICE_BLOCK_SIZE: return "invalid device block size";
    case ERROR_READ_DEVICE: return "read device fail";
    case ERROR_WRITE_DEVICE: return "write device fail";
    case ERROR_PARSE_DEVICE_LIST: return "error parsing device list";
    case ERROR_END_OF_ARCHIVE: return "end of archive";
    case ERROR_NO_FILE_ENTRY: return "no file entry";
    case ERROR_NO_FILE_DATA: return "no file data entry";
    case ERROR_NO_DIRECTORY_ENTRY: return "no directory entry";
    case ERROR_NO_LINK_ENTRY: return "no link entry";
    case ERROR_NO_SPECIAL_ENTRY: return "no special entry";
    case ERROR_NO_IMAGE_ENTRY: return "no image entry";
    case ERROR_NO_IMAGE_DATA: return "no image data entry";
    case ERROR_END_OF_DATA: return "end of data";
    case ERROR_CRC_ERROR: return "CRC error";
    case ERROR_FILE_INCOMPLETE: return "file is incomplete";
    case ERROR_WRONG_FILE_TYPE: return "wrong file type";
    case ERROR_FILES_DIFFER: return "files differ";
    case ERROR_CORRUPT_DATA: return "corrupt data or invalid password";
    case ERROR_NOT_AN_INCREMENTAL_FILE: return "invalid incremental file";
    case ERROR_WRONG_INCREMENTAL_FILE_VERSION: return "wrong incremental file version";
    case ERROR_CORRUPT_INCREMENTAL_FILE: return "corrupt incremental file";
    case ERROR_HOST_NOT_FOUND: return "host not found";
    case ERROR_CONNECT_FAIL:
      {
        strncpy(errorText,strerror(ERRNO),sizeof(errorText)-1); errorText[sizeof(errorText)-1] = '\0';
        if (ERROR_TEXT != NULL)
        {
          strncat(errorText,": ",sizeof(errorText)-1-strlen(errorText));
          strncat(errorText,ERROR_TEXT,sizeof(errorText)-1-strlen(errorText));
        }
      }
      break;
    case ERROR_NO_LOGIN_NAME: return "no login name given";
    case ERROR_NO_PASSWORD: return "no password given";
    case ERROR_NETWORK_SEND: return "sending data fail";
    case ERROR_NETWORK_RECEIVE: return "receiving data fail";
    case ERROR_NETWORK_EXECUTE_FAIL: return "execute command fail";
    case ERROR_NO_RESPONSE: return "no response from remote program";
    case ERROR_INVALID_RESPONSE: return "invalid response from remote program";
    case ERROR_INVALID_DEVICE_SPECIFIER: return "invalid device specifier";
    case ERROR_LOAD_VOLUME_FAIL: return "load volume fail";
    case ERROR_FORK_FAIL: return "fork for execute external program fail";
    case ERROR_EXEC_FAIL:
      {
        strncpy(errorText,"execute external program fail",sizeof(errorText)-1); errorText[sizeof(errorText)-1] = '\0';
        if (ERROR_TEXT != NULL)
        {
          strncat(errorText,": ",sizeof(errorText)-1-strlen(errorText));
          strncat(errorText,ERROR_TEXT,sizeof(errorText)-1-strlen(errorText));
        }
      }
      break;
    case ERROR_PARSING: return "error parsing data";
    case ERROR_EXPECTED_PARAMETER: return "expected parameter";
    case ERROR_UNKNOWN_VALUE: return "unknown value";
    case ERROR_AUTHORIZATION: return "authorization error";
    case ERROR_JOB_NOT_FOUND: return "job not found";
    case ERROR_JOB: return "job error";

  }

  return errorText;
}
